#include "postgres.h"
#include "access/heapam.h"
#include "catalog/namespace.h"
#include "utils/builtins.h"

Datum	size_rules(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(size_rules);

Datum
size_rules(PG_FUNCTION_ARGS)
{
	text	   	*relname = PG_GETARG_TEXT_P(0);
	
	RangeVar   	*relrv;
	Relation	rel;
	HeapScanDesc 	scan;
	HeapTuple	tuple;
	TupleDesc	tupdesc;
	
	int 		size=0;
	int 		cont;
	AttrNumber 	numatt;
	Datum		attr;
	bool		isnull;
	
	static int32 ip=0;
	static int32 it=0;
	
	relrv = makeRangeVarFromNameList(textToQualifiedNameList(relname,"size_rules"));
	rel = heap_openrv(relrv, AccessShareLock);
	tupdesc = rel->rd_att;
		
	scan = heap_beginscan(rel, SnapshotAny, 0, NULL);
	heap_restrpos(scan);
	
	if(ip == 0 && it == 0){
		while ((tuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
			it++;
	}
	
	for(cont = 0; cont <= ip-1; cont++){
		tuple = heap_getnext(scan, ForwardScanDirection);
	}
		
	while ((tuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
	{	
		size = 0;		
		for(numatt = 1 ; numatt < tupdesc->natts ; numatt ++)
		{
			attr = heap_getattr(tuple,numatt,tupdesc,&isnull);
			if (!isnull)
				size++;
		}
		break;
	}
	heap_endscan(scan);
	heap_close(rel, AccessShareLock);

	if(ip == it-1){
		ip=0;
		it=0;	
	}
	else
		ip++;
	
	PG_RETURN_INT64(size);
}


