/*-------------------------------------------------------------------------
 *
 * nodeAssociator.c
 *	  Soporta las rutinas para Associator.
 *
 * Portions Copyright (c) 2005, PostgreSQL GRIAS Development Group
 * Portions Copyright (c) 2005, Regents of the University of Nariño
 *
 *
 * IDENTIFICATION
 *	  $Header: /cvsroot/pgsql-server/src/backend/executor/nodeSeqscan.c,v 1.37 2002/09/04 20:31:18 momjian Exp $
 *
 *-------------------------------------------------------------------------
 */
/*
 * INTERFACE ROUTINES
 *		ExecAssociator			Retorna las tuplas generadas por Asociator.
 *		ExecEndSeqScan			Finaliza cualquier almacenamiento.
 *
 */
#include "postgres.h"
#include "access/heapam.h"
#include "executor/execdebug.h"
#include "executor/nodeAssociator.h"
#include "parser/parsetree.h"
#include "utils/tuplestore.h"

static void controlPorden(int desde, int cuantos, int *controlPosicion, int donde);
static void GenerarAssociator(HeapTuple	heapTuple,
							  Tuplestorestate *tuplestorestate,
							  AssociatorState *astate);
		

static void controlPorden(int desde, int cuantos, int *controlPosicion, int donde) {
	int i,j = donde;
	for (i = 0; i < cuantos - donde; i++) {
		controlPosicion[j] = desde;
		desde++;
		j++;
	}
}

/* ----------------------------------------------------------------
 *		GenerarAssociator(node)
 *
 *		Algoritmo que se encargua de recibir una tupla y generar
 * 		las diferentes asociaciones de ésa tupla y almacenarlas en 
 * 		memoria.
 *
 */
static void GenerarAssociator(HeapTuple heapTuple,
							  Tuplestorestate *tuplestorestate,
							  AssociatorState *astate) 
{
	HeapTuple	ht;
	int controlPosicion[255];
	int active, LocalSet, i, Attribs = astate->tupla->ttc_tupleDescriptor->natts;
		
	ht = heapTuple;
	astate->num_tuplas = 0;
	
	Datum		at[Attribs];
	char		nulls[Attribs];
	bool		isnull;
		
	for (LocalSet = astate->rginicio; LocalSet <= astate->rgfin; LocalSet++) { //1
			active = 0;
			controlPorden(0, LocalSet, controlPosicion, 0);
			active = LocalSet - 1;
			//cami--begin
			while (controlPosicion[0] < Attribs - (LocalSet - 1)) { //2
				
				while (controlPosicion[active] < Attribs) { //3
					int flag=1;		
										
					for (i = 0; i < Attribs; i++) {
						nulls[i] = 'n';
					}

					for (i = 0; i < LocalSet; i++) {
						at[controlPosicion[i]] = heap_getattr(ht, controlPosicion[i]+1, astate->tupla->ttc_tupleDescriptor, &isnull);
						if(!at[controlPosicion[i]])//SI ES NULO
						{	flag=0;
							break;
						}					
						nulls[controlPosicion[i]] = ' ';
					}
					if(flag==1){					//SI NO ES NULO ALMACENA LA TUPLA		
						heapTuple = heap_formtuple(astate->tupla->ttc_tupleDescriptor,at,nulls);
						tuplestore_puttuple(tuplestorestate,(void *) heapTuple);
						astate->num_tuplas=astate->num_tuplas + 1;
					}	
					controlPosicion[active] = controlPosicion[active] + 1;
				} //3
				
				if (active > 0) { //4
					controlPosicion[active] = controlPosicion[active] - 1;
					//arregla el paso mas del while anterior
					while (active > 0 && !(controlPosicion[active - 1] + 1 < controlPosicion[active])) {//5
						active = active - 1;
					} //5
					if (active > 0) { //6
						active = active - 1;
					} //6
					//cami --
					controlPorden(controlPosicion[active] + 1, LocalSet, controlPosicion, active);
					active = LocalSet - 1;
				} //4
			} //2
		} //1
	tuplestore_donestoring(tuplestorestate);
}


/* ----------------------------------------------------------------
 *		ExecAssociator(node)
 *
 *		Recibe una tupla y entregua las tuplas generadas por el algoritmo
 * 		Associator.
 *		Llama al nodo SeqScan el cual retorna la tupla
 *		a la cual se le aplica el algoritmo Associator
 *
 */

TupleTableSlot *
ExecAssociator(Associator *node)
{	
	EState	   *estate;
	AssociatorState *astate;
	Tuplestorestate *tuplestorestate;
	TupleTableSlot *slot;
	HeapTuple	heapTuple;
	TupleDesc	tupdesc;	
	bool should_free;
	/*
	 * get state info from node
	 */
	astate = node->astate;
	estate = node->plan.state;
	tuplestorestate = (Tuplestorestate *) astate->tuplestorestate;	
lnext3:;		
	if (astate->pedir_tupla)
	{
		Plan	   *outerNode;
		
		outerNode = outerPlan((Plan *) node);
		tupdesc = ExecGetTupType(outerNode);//tupdesc = slot->ttc_tupleDescriptor;
		slot = ExecProcNode(outerNode, (Plan *) node );
				
		if (TupIsNull(slot))
			return NULL;
			
		ExecAssignResultType(&astate->csstate.cstate, tupdesc, false);
		astate->tupla = slot;
		astate->pedir_tupla=false;
		
		goto lnext1;
	}
	else
	{
		slot =  astate->tupla;
lnext1:;		
		heapTuple = astate->tupla->val;
		
		if (!astate->retornar_tupla) 
		{
			tuplestorestate = tuplestore_begin_heap(true,1024);//sizeof(HeapTuple)
			astate->tuplestorestate = (void *)tuplestorestate;
			GenerarAssociator(heapTuple,tuplestorestate,astate);
			astate->retornar_tupla=true;	

			goto lnext2;		
		}
		else 
		{	
lnext2:;				
			if (astate->out_tuplas < astate->num_tuplas)
				{
					heapTuple = tuplestore_getheaptuple(tuplestorestate,true,&should_free);					
					astate->out_tuplas = astate->out_tuplas + 1;
				}
			else
				{
				
					tuplestore_end((Tuplestorestate *) astate->tuplestorestate );
					astate->tuplestorestate = NULL;					
					astate->retornar_tupla = false;
					astate->pedir_tupla = true;//OJO NO QUITAR O SE QUEDA EN UN LOO...OUPS
					astate->out_tuplas = 0;
					goto lnext3;
				}
		}
	}

	return ExecStoreTuple(heapTuple, slot, InvalidBuffer, should_free);
}



/* ----------------------------------------------------------------
 *		ExecInitAssociator
 * ----------------------------------------------------------------
 */
bool
ExecInitAssociator(Associator *node, EState *estate, Plan *parent)
{
	AssociatorState *astate;
	Plan			*outerPlan;
	
	/*
	 * assign the node's execution state
	 */
	node->plan.state = estate;
	
	/*
	 * create new CommonScanState for node
	 */
	astate = makeNode(AssociatorState);
	astate->out_tuplas = 0;
	astate->num_tuplas = 0;
	astate->rginicio = node->inicio;
	astate->rgfin = node->fin;
	astate->pedir_tupla=true;
	astate->retornar_tupla=false;
	astate->tuplestorestate=NULL;
	node->astate = astate;
	
#define ASSOCIATOR_NSLOTS 2
	
	/*
	 * tuple table initialization
	 */
	ExecInitResultTupleSlot(estate, &astate->csstate.cstate);
	ExecInitScanTupleSlot(estate, &astate->csstate);
	
	/*
	 * initializes child nodes
	 */
	outerPlan = outerPlan((Plan *) node);//outerPlan(node);
	ExecInitNode(outerPlan, estate, (Plan *) node);	
	
	/*
	 * initialize tuple type.
	 */
	ExecAssignResultTypeFromOuterPlan((Plan *) node, &astate->csstate.cstate);
	ExecAssignScanTypeFromOuterPlan((Plan *) node, &astate->csstate);
	
	astate->csstate.cstate.cs_ProjInfo=NULL;
	
	return TRUE;
}

int
ExecCountSlotsAssociator(Associator *node)
{
	return ExecCountSlotsNode(outerPlan(node)) +
		ExecCountSlotsNode(innerPlan(node)) +
		ASSOCIATOR_NSLOTS;
}

/* ----------------------------------------------------------------
 *		ExecEndAssociator
 *
 *		Libera cualquier almacenamiento.
 * ----------------------------------------------------------------
 */
void
ExecEndAssociator(Associator *node)
{
	AssociatorState *astate;
	Plan	   *outerPlan;

	/*
	 * get information from node
	 */
	astate = node->astate;
	
	/*
	 * shut down the subplan
	 */
	outerPlan = outerPlan((Plan *) node);
	ExecEndNode(outerPlan, (Plan *) node);
	
	/*
	 * clean out the tuple table
	 */
	ExecClearTuple(astate->csstate.css_ScanTupleSlot);
	
	/*
	 * Release tuplestore resources
	 */
	if (astate->tuplestorestate != NULL)
		tuplestore_end((Tuplestorestate *) astate->tuplestorestate );
	astate->tuplestorestate = NULL;

	pfree(astate);
	node->astate = NULL;	
}
