/*-------------------------------------------------------------------------
 *
 * nodeEquiKeep.c
 *	  Soporta las rutinas para Keep.
 *
 * Portions Copyright (c) 2005, PostgreSQL GRIAS Development Group
 * Portions Copyright (c) 2005, Regents of the University of Nariño
 *
 *
 * IDENTIFICATION
 *	  $Header: /src/backend/executor/nodeEquiKeep.c,v 1.0 2005/03/20 Camiloman!
 *
 *-------------------------------------------------------------------------
 */
/*
 * INTERFACE ROUTINES
 *		ExecEquiKeep retorna la tupla generada por Keep.
 *		ExecEndEquiKeep			Finaliza cualquier almacenamiento.
 *
 */
#include "postgres.h"

#include "access/heapam.h"
#include "executor/execdebug.h"
#include "executor/nodeEquiKeep.h"
#include "parser/parsetree.h"
#include "utils/tuplestore.h"

/* ----------------------------------------------------------------
 *		ExecEquiKeep(node)
 *
 *		Recibe una tupla y entregua la tupla conservando solo el valor que
 * 		el Keep marca como verdadero.
 *		Llama al nodo SeqScan el cual retorna la tupla
 *		a la cual se le aplica el roceso de KEEP
 *
 */

TupleTableSlot *
ExecEquiKeep(EquiKeep *node)
{	
	EState	   *estate;
	EquiKeepState *kstate;
	TupleTableSlot *slot;
	HeapTuple	heapTuple;
	TupleDesc	tupdesc;	
	Plan	   *outerNode;
	
	ExprContext *econtext;
	List *lst;
	int i,j,k;
	bool rta;
		
	/*
	 * get state info from node
	 */
	kstate = node->kstate;
	estate = node->plan.state;

	outerNode = outerPlan((Plan *) node);
	tupdesc = ExecGetTupType(outerNode);//tupdesc = slot->ttc_tupleDescriptor;
	
	Datum		at[tupdesc->natts];
	char		nulls[tupdesc->natts];
	bool		isnull;
	
lnext1:;	
	slot = ExecProcNode(outerNode, (Plan *) node );
			
	if (TupIsNull(slot))
		return NULL;
			
	econtext = kstate->csstate.cstate.cs_ExprContext;
	ResetExprContext(econtext);
	econtext->ecxt_scantuple = slot;
	heapTuple= slot->val;
	rta = false;
	k=0;
	i=0;
		
	foreach(lst, kstate->kqual)
	{
		at[i] = heap_getattr(heapTuple, i+1, tupdesc , &isnull);
		Node	*tkqual = (Node *) lfirst(lst);
		rta = ExecQual((List *)tkqual, econtext, false);
		if(rta) 
			nulls[i] = ' ';
		else{	
			nulls[i] = 'n';
			k++;
		}
		i++;
	}	
	for(j=i;j<=tupdesc->natts;j++){
		at[j] = heap_getattr(heapTuple, j+1, tupdesc , &isnull);
		nulls[j]='n';
		k++;
	}
	
	if(k > tupdesc->natts)
		goto lnext1;//Si todos son nulos, pido otra tupla
	
	heapTuple = heap_formtuple(tupdesc,at,nulls);
					
	ExecAssignResultType(&kstate->csstate.cstate, tupdesc, false);
	slot=  kstate->csstate.cstate.cs_ResultTupleSlot;
	
	return ExecStoreTuple(heapTuple, slot, InvalidBuffer, false);
}



/* ----------------------------------------------------------------
 *		ExecInitEquiKeep
 * ----------------------------------------------------------------
 */
bool
ExecInitEquiKeep(EquiKeep *node, EState *estate, Plan *parent)
{
	EquiKeepState *kstate;
	Plan			*outerPlan;
	
	/*
	 * assign the node's execution state
	 */
	node->plan.state = estate;
	
	/*
	 * create new CommonScanState for node
	 */
	kstate = makeNode(EquiKeepState);
	kstate->kqual = node->kqual;
	node->kstate = kstate;
	
	ExecAssignExprContext(estate, &kstate->csstate.cstate);//Yo la metí
	
#define EQUIKEEP_NSLOTS 2
	
	/*
	 * tuple table initialization
	 */
	ExecInitResultTupleSlot(estate, &kstate->csstate.cstate);
	ExecInitScanTupleSlot(estate, &kstate->csstate);
	 
	/*
	 * initializes child nodes
	 */
	outerPlan = outerPlan((Plan *) node);//outerPlan(node);
	ExecInitNode(outerPlan, estate, (Plan *) node);	
	
	/*
	 * initialize tuple type.
	 */
	ExecAssignResultTypeFromOuterPlan((Plan *) node, &kstate->csstate.cstate);
	ExecAssignScanTypeFromOuterPlan((Plan *) node, &kstate->csstate);
	kstate->csstate.cstate.cs_ProjInfo=NULL;
	
	return TRUE;
}

int
ExecCountSlotsEquiKeep(EquiKeep *node)
{
	return ExecCountSlotsNode(outerPlan(node)) +
		ExecCountSlotsNode(innerPlan(node)) +
		EQUIKEEP_NSLOTS;
}

/* ----------------------------------------------------------------
 *		ExecEndEquiKeep
 *
 *		Libera cualquier almacenamiento.
 * ----------------------------------------------------------------
 */
void
ExecEndEquiKeep(EquiKeep *node)
{
	EquiKeepState *kstate;
	Plan	   *outerPlan;

	/*
	 * get information from node
	 */
	kstate = node->kstate;
	
	/*
	 * shut down the subplan
	 */
	outerPlan = outerPlan((Plan *) node);
	ExecEndNode(outerPlan, (Plan *) node);
	
	/*
	 * clean out the tuple table
	 */
	ExecClearTuple(kstate->csstate.css_ScanTupleSlot);
	
	pfree(kstate);
	node->kstate = NULL;
}
