/*-------------------------------------------------------------------------
 *
 * nodeMate.c
 *	  Support routines for Mate Clause.
 *
 * Portions Copyright (c) 2005, PostgreSQL Muerguanos Development Group
 * Portions Copyright (c) 2005, Regents of the University of Nariño
 *
 *
 * IDENTIFICATION
 *	  $Header: /cvsroot/pgsql-server/src/backend/executor/nodeSeqscan.c,v 1.37 2002/09/04 20:31:18 momjian Exp $
 *
 *-------------------------------------------------------------------------
 */
/*
 * INTERFACE ROUTINES
 *		ExecMate				Ejecuta el algoritmo Mate.
 *		ExecEndSeqScan			Termina el nodo Mate.
 */
#include "postgres.h"
#include "access/heapam.h"
#include "executor/execdebug.h"
#include "executor/nodeMate.h"
#include "parser/parsetree.h"
#include "utils/tuplestore.h"

static void controlPorden(int desde, int cuantos, int *controlPosicion, int donde);
static void GenerarMate(HeapTuple	heapTuple,
							  Tuplestorestate *tuplestorestate,
							  MateState *matestate);
		

static void controlPorden(int desde, int cuantos, int *controlPosicion, int donde) {
	int i,j = donde;
	for (i = 0; i < cuantos - donde; i++) {
		controlPosicion[j] = desde;
		desde++;
		j++;
	}
}
/* ----------------------------------------------------------------
 *		GenerarMate(node)
 *
 *		Algoritmo que se encargua de recibir una tupla y generar
 * 		las diferentes asociaciones de ésa tupla teniendo en cuenta
 * 		los atributos condicion con el atributo clase y almacenarlas en 
 * 		memoria.
 *
 */
static void GenerarMate(HeapTuple heapTuple,
							  Tuplestorestate *tuplestorestate,
							  MateState *matestate) 
{
	HeapTuple	ht;
	int controlPosicion[255];
	int active, LocalSet, i, Attribs = matestate->tupla->ttc_tupleDescriptor->natts-1;
		
	ht = heapTuple;
	matestate->num_tuplas = 0;
	
	Datum		at[Attribs];
	char		nulls[Attribs];
	bool		isnull;
	
	for (LocalSet = 1; LocalSet <= Attribs; LocalSet++) { //1
			active = 0;
			controlPorden(0, LocalSet, controlPosicion, 0);
			active = LocalSet - 1;
			//cami--begin
			while (controlPosicion[0] < Attribs - (LocalSet - 1)) { //2
				while (controlPosicion[active] < Attribs) { //3
					int flag=1;		
										
					for (i = 0; i < Attribs; i++) {
						nulls[i] = 'n';
					}
					
					for (i = 0; i < LocalSet; i++) {
						at[controlPosicion[i]] = heap_getattr(ht, controlPosicion[i]+1, matestate->tupla->ttc_tupleDescriptor, &isnull);
						if(!at[controlPosicion[i]])//SI ES NULO
						{	flag=0;
							break;
						}					
						nulls[controlPosicion[i]] = ' ';
					}
//					if(flag==1){					//SI NO ES NULO ALMACENA LA TUPLA	
//						at[Attribs] = heap_getattr(ht, Attribs+1, matestate->tupla->ttc_tupleDescriptor, &isnull);
//						nulls[Attribs] = ' ';	
//						heapTuple = heap_formtuple(matestate->tupla->ttc_tupleDescriptor,at,nulls);
//						tuplestore_puttuple(tuplestorestate,(void *) heapTuple);
//						matestate->num_tuplas=matestate->num_tuplas + 1;
//					}	
					
					if(flag==1){					//SI NO ES NULO ALMACENA LA TUPLA	
						at[Attribs] = heap_getattr(ht, Attribs+1, matestate->tupla->ttc_tupleDescriptor, &isnull);
						nulls[Attribs] = ' ';	
						heapTuple = heap_formtuple(matestate->tupla->ttc_tupleDescriptor,at,nulls);
						tuplestore_puttuple(tuplestorestate,(void *) heapTuple);
						matestate->num_tuplas = matestate->num_tuplas + 1;
					}
					
					
					controlPosicion[active] = controlPosicion[active] + 1;
				} //3
				
				if (active > 0) { //4
					controlPosicion[active] = controlPosicion[active] - 1;
					//arregla el paso mas del while anterior
					while (active > 0 && !(controlPosicion[active - 1] + 1 < controlPosicion[active])) {//5
						active = active - 1;
					} //5
					if (active > 0) { //6
						active = active - 1;
					} //6
					//cami --
					controlPorden(controlPosicion[active] + 1, LocalSet, controlPosicion, active);
					active = LocalSet - 1;
				} //4
			} //2
		} //1
	tuplestore_donestoring(tuplestorestate);
}

/* ----------------------------------------------------------------
 *		ExecMate(node)
 *		Llama a las rutinas para ejecucion del algoritmo Mate
 * 		e inicializa las estructuras necesarias.
 */

TupleTableSlot *
ExecMate(Mate *node)
{	
	EState	   *estate;
	MateState *matestate;
	Tuplestorestate *tuplestorestate;
	TupleTableSlot *slot;
	HeapTuple	heapTuple;
	TupleDesc	tupdesc;	
	bool should_free;
	/*
	 * get state info from node
	 */
	matestate = node->matestate;
	estate = node->plan.state;
	tuplestorestate = (Tuplestorestate *) matestate->tuplestorestate;	
	
	if (matestate->pedir_tupla)
	{
		Plan	   *outerNode;
		
		outerNode = outerPlan((Plan *) node);
		tupdesc = ExecGetTupType(outerNode);//tupdesc = slot->ttc_tupleDescriptor;
		slot = ExecProcNode(outerNode, (Plan *) node );
				
		if (TupIsNull(slot))
			return NULL;
			
		ExecAssignResultType(&matestate->csstate.cstate, tupdesc, false);
		matestate->tupla = slot;
		matestate->pedir_tupla=false;
		
		goto lnext1;
	}
	else
	{
		slot =  matestate->tupla;
lnext1:;		
		heapTuple = matestate->tupla->val;
		
		if (!matestate->retornar_tupla) 
		{
			tuplestorestate = tuplestore_begin_heap(true,1024);//sizeof(HeapTuple)
			matestate->tuplestorestate = (void *)tuplestorestate;
			GenerarMate(heapTuple,tuplestorestate,matestate);
			matestate->retornar_tupla=true;	

			goto lnext2;		
		}
		else 
		{	
lnext2:;				
			if (matestate->out_tuplas < matestate->num_tuplas-1)
				{
					heapTuple = tuplestore_getheaptuple(tuplestorestate,true,&should_free);					
					matestate->out_tuplas = matestate->out_tuplas + 1;
				}
			else
				{
				heapTuple = tuplestore_getheaptuple(tuplestorestate,true,&should_free);					
				tuplestore_end((Tuplestorestate *) matestate->tuplestorestate );
				matestate->tuplestorestate = NULL;
				matestate->retornar_tupla = false;
				matestate->pedir_tupla = true;//OJO NO QUITAR O SE QUEDA EN UN LOO...OUPS
				matestate->out_tuplas = 0;
				}
		}
	}

	return ExecStoreTuple(heapTuple, slot, InvalidBuffer, should_free);
}


/* ----------------------------------------------------------------
 *		ExecInitMate
 * ----------------------------------------------------------------
 */
bool
ExecInitMate(Mate *node, EState *estate, Plan *parent)
{
	MateState *matestate;
	Plan			*outerPlan;
	
	/*
	 * assign the node's execution state
	 */
	node->plan.state = estate;
	
	/*
	 * create new CommonScanState for node
	 */
	matestate = makeNode(MateState);
	matestate->mateList = node->mateList;
	matestate->out_tuplas = 0;
	matestate->num_tuplas = 0;
	matestate->pedir_tupla=true;
	matestate->retornar_tupla=false;
	matestate->tuplestorestate=NULL;
	node->matestate = matestate;
		
#define MATE_NSLOTS 2
	
	/*
	 * tuple table initialization
	 */
	ExecInitResultTupleSlot(estate, &matestate->csstate.cstate);
	ExecInitScanTupleSlot(estate, &matestate->csstate);
	
	/*
	 * initializes child nodes
	 */
	outerPlan = outerPlan((Plan *) node);//outerPlan(node);
	ExecInitNode(outerPlan, estate, (Plan *) node);	
	
	/*
	 * initialize tuple type.
	 */
	ExecAssignResultTypeFromOuterPlan((Plan *) node, &matestate->csstate.cstate);
	ExecAssignScanTypeFromOuterPlan((Plan *) node, &matestate->csstate);
	
	matestate->csstate.cstate.cs_ProjInfo=NULL;
	
	return TRUE;
}

int
ExecCountSlotsMate(Mate *node)
{
	return ExecCountSlotsNode(outerPlan(node)) +
		ExecCountSlotsNode(innerPlan(node)) +
		MATE_NSLOTS;
}

/* ----------------------------------------------------------------
 *		ExecEndMate
 *
 *		frees any storage allocated through C routines.
 * ----------------------------------------------------------------
 */
void
ExecEndMate(Mate *node)
{
	MateState *matestate;
	Plan	   *outerPlan;

	/*
	 * get information from node
	 */
	matestate = node->matestate;
	
	/*
	 * shut down the subplan
	 */
	outerPlan = outerPlan((Plan *) node);
	ExecEndNode(outerPlan, (Plan *) node);
	
	/*
	 * clean out the tuple table
	 */
	ExecClearTuple(matestate->csstate.css_ScanTupleSlot);
	
	/*
	 * Release tuplestore resources
	 */
	if (matestate->tuplestorestate != NULL)
		tuplestore_end((Tuplestorestate *) matestate->tuplestorestate );
	matestate->tuplestorestate = NULL;
	
	
	pfree(matestate);
	node->matestate = NULL;	
}
